/*
 *	$Id: $
 */

#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "miscadmin.h"
#include "access/genam.h"
#include "access/heapam.h"
#include "access/nbtree.h"
#include "storage/bufmgr.h"
#include "storage/lmgr.h"
#include "storage/pmsignal.h"

PG_FUNCTION_INFO_V1(btree_stats);

Datum
btree_stats(PG_FUNCTION_ARGS)
{
	/* TODO */
	PG_RETURN_NULL();
}



/*
 *	_bt_pgaddtup() -- add a tuple to a particular page in the index.
 *
 * XXX copied from backend\access\nbtree\nbtinsert.c
 */
static void
_bt_pgaddtup(Relation rel,
			 Page page,
			 Size itemsize,
			 BTItem btitem,
			 OffsetNumber itup_off,
			 const char *where)
{
	BTPageOpaque opaque = (BTPageOpaque) PageGetSpecialPointer(page);
	BTItemData	truncitem;

	if (!P_ISLEAF(opaque) && itup_off == P_FIRSTDATAKEY(opaque))
	{
		memcpy(&truncitem, btitem, sizeof(BTItemData));
		truncitem.bti_itup.t_info = sizeof(BTItemData);
		btitem = &truncitem;
		itemsize = sizeof(BTItemData);
	}

	if (PageAddItem(page, (Item) btitem, itemsize, itup_off,
					LP_USED) == InvalidOffsetNumber)
		elog(PANIC, "failed to add item to the %s for \"%s\"",
			 where, RelationGetRelationName(rel));
}


/*
 * We fetch the metapage and sanity check it ourselves to make
 * sure the rest of the format knowledge here is valid.  If we
 * just called _bt_getroot() that code and disk index format may
 * match each other but not us.
 */
static void
check_btree_index(Relation rel)
{
	Buffer		metabuf;
	Page		metapg;
	BTPageOpaque metaopaque;
	BTMetaPageData *metad;

	metabuf = _bt_getbuf(rel, BTREE_METAPAGE, BT_READ);
	metapg = BufferGetPage(metabuf);
	metaopaque = (BTPageOpaque) PageGetSpecialPointer(metapg);
	metad = BTPageGetMeta(metapg);

	/* sanity-check the metapage */
	if (!(metaopaque->btpo_flags & BTP_META) ||
		metad->btm_magic != BTREE_MAGIC)
		ereport(ERROR,
				(errcode(ERRCODE_INDEX_CORRUPTED),
				 errmsg("index \"%s\" is not a btree",
						RelationGetRelationName(rel))));

	if (metad->btm_version != BTREE_VERSION)
		ereport(ERROR,
				(errcode(ERRCODE_INDEX_CORRUPTED),
				 errmsg("version mismatch in index \"%s\": file version %d, code version %d",
						RelationGetRelationName(rel),
						metad->btm_version, BTREE_VERSION)));

	_bt_relbuf(rel, metabuf);		/* done with the meta page */
	return;
}

#if 0
static void
logPageInfo(Relation iRel, BlockNumber blkno, Buffer buf,
		    Page page, BTPageOpaque opaque, const char *desc)
{
	Size	space, pgsize;
	double	full;
	int		items;

	items = PageGetMaxOffsetNumber(page) - P_FIRSTDATAKEY(opaque);
	pgsize = PageGetPageSize(page);
	space = PageGetFreeSpace(page);
	full = 100.0 - ((double) space * 100.0 / pgsize);

	ereport(LOG,
			(errcode(ERRCODE_WARNING),
			 errmsg("%-5s block %4u has %3u items, %4u/%u free = %4.1f%% full",
					desc, blkno, items, space, pgsize, full)));
}
#else
#define logPageInfo(rel,blk,buf,page,op,desc)
#endif


/*
 * Returns the amount of free space that would be left in the
 * right page if all items from the left where moved into it.
 * Can be negative if all items won't fit.
 */
static int
mergedFreeSpace(Page lPage, Page rPage)
{
	ItemId	itemid;
	int		space, lData;

	/* Total free space available on a btree page, after fixed overhead */
	space = PageGetPageSize(lPage) - SizeOfPageHeaderData -
			MAXALIGN(sizeof(BTPageOpaqueData));
	lData = space - (int) PageGetFreeSpace(lPage);

	/* The left page's high key is not moved */
	itemid = PageGetItemId(lPage, P_HIKEY);
	lData -= (int) (MAXALIGN(ItemIdGetLength(itemid)) + sizeof(ItemIdData));

	space = PageGetFreeSpace(rPage) - (int) MAXALIGN(lData);
	return space;
}


static int
mergeBuffers(Relation iRel, Buffer lbuf, Buffer rbuf)
{
	Page		lPage, rPage;
	Page		olPage, orPage;
	BTPageOpaque lOpaque, rOpaque;
	BTPageOpaque olOpaque, orOpaque;
	OffsetNumber i, off, maxoff;
	BTItem		item;
	ItemId		itemid;
	Size		itemsz;

	/* Try and get write locks */
	LockBuffer(lbuf, BUFFER_LOCK_UNLOCK);	/* XXX */
	if (!ConditionalLockBuffer(lbuf))
	{
		ereport(INFO,
				(errcode(ERRCODE_WARNING),
				 errmsg("failed to get write lock on lblock %u",
					    BufferGetBlockNumber(lbuf))));
		return false;
	}
	LockBuffer(rbuf, BUFFER_LOCK_UNLOCK);	/* XXX */
	if (!ConditionalLockBuffer(rbuf))
	{
		ereport(INFO,
				(errcode(ERRCODE_WARNING),
				 errmsg("failed to get write lock on rblock %u",
					    BufferGetBlockNumber(rbuf))));
		return false;
	}

	olPage = BufferGetPage(lbuf);
	orPage = BufferGetPage(rbuf);
	olOpaque = (BTPageOpaque) PageGetSpecialPointer(olPage);
	orOpaque = (BTPageOpaque) PageGetSpecialPointer(orPage);

	lPage = PageGetTempPage(olPage, sizeof(BTPageOpaqueData));
	rPage = PageGetTempPage(orPage, sizeof(BTPageOpaqueData));
	_bt_pageinit(lPage, BufferGetPageSize(lbuf));
	_bt_pageinit(rPage, BufferGetPageSize(rbuf));
	lOpaque = (BTPageOpaque) PageGetSpecialPointer(lPage);
	rOpaque = (BTPageOpaque) PageGetSpecialPointer(rPage);

	/* XXX is this needed? */
	lOpaque->btpo_flags = olOpaque->btpo_flags; // | BTP_HALF_DEAD;
	rOpaque->btpo_flags = orOpaque->btpo_flags;
	lOpaque->btpo_prev = olOpaque->btpo_prev;
	rOpaque->btpo_prev = orOpaque->btpo_prev;
	lOpaque->btpo_next = olOpaque->btpo_next;
	rOpaque->btpo_next = orOpaque->btpo_next;
	lOpaque->btpo.level = olOpaque->btpo.level;
	rOpaque->btpo.level = orOpaque->btpo.level;

	off = P_HIKEY;

	/* Copy right highkey (if one) */
	if (!P_RIGHTMOST(orOpaque))
	{
		itemid = PageGetItemId(orPage, P_HIKEY);
		itemsz = ItemIdGetLength(itemid);
		item = (BTItem) PageGetItem(orPage, itemid);

#if 0
		ereport(LOG,
				(errcode(ERRCODE_WARNING),
				 errmsg("Copying hikey from right to merged")));
#endif
		if (PageAddItem(rPage, (Item) item, itemsz, off,
						LP_USED) == InvalidOffsetNumber)
			elog(PANIC, "failed to add hikey to the merge page");
		off = OffsetNumberNext(off);
	}

	/* Set new left high key */
	itemid = PageGetItemId(olPage, P_FIRSTDATAKEY(olOpaque));
	itemsz = ItemIdGetLength(itemid);
	item = (BTItem) PageGetItem(olPage, itemid);
	if (PageAddItem(lPage, (Item) item, itemsz, P_HIKEY,
					LP_USED) == InvalidOffsetNumber)
		elog(PANIC, "failed to add hikey to the new left");

	/* Now copy all the data items from the left page */
	maxoff = PageGetMaxOffsetNumber(olPage);
	for (i = P_FIRSTDATAKEY(olOpaque); i <= maxoff;
		 i = OffsetNumberNext(i))
	{
		itemid = PageGetItemId(olPage, i);
		itemsz = ItemIdGetLength(itemid);
		item = (BTItem) PageGetItem(olPage, itemid);

		_bt_pgaddtup(iRel, rPage, itemsz, item, off, "left page merge");
		off = OffsetNumberNext(off);
	}

	/* Now copy all the data items from the right page */
	maxoff = PageGetMaxOffsetNumber(orPage);
	for (i = P_FIRSTDATAKEY(orOpaque); i <= maxoff;
		 i = OffsetNumberNext(i))
	{
		itemid = PageGetItemId(orPage, i);
		itemsz = ItemIdGetLength(itemid);
		item = (BTItem) PageGetItem(orPage, itemid);

		_bt_pgaddtup(iRel, rPage, itemsz, item, off, "left page merge");
		off = OffsetNumberNext(off);
	}

	START_CRIT_SECTION();

	/* XLOG stuff */
	if (!iRel->rd_istemp)
	{
		/* XXX for now just warn */
		static int warned = 0;
		if (!warned)
		{
			ereport(WARNING,
					(errcode(ERRCODE_WARNING),
					 errmsg("*** Not recording XLOG entry for btree index page merges!!")));
			warned = 1;
		}
	}

	PageRestoreTempPage(lPage, olPage);
	PageRestoreTempPage(rPage, orPage);
	END_CRIT_SECTION();

#if 0
	_bt_wrtnorelbuf(iRel, lbuf);
	_bt_wrtnorelbuf(iRel, rbuf);
#endif
	return true;
}


/* instead of a limit on number of blocks to process
 * it should be time based or work based */

PG_FUNCTION_INFO_V1(btree_index_compress);

Datum
btree_index_compress(PG_FUNCTION_ARGS)
{
	Oid			indexId = PG_GETARG_OID(0);
	float4		fillpct = PG_GETARG_FLOAT4(1);
	BlockNumber	startblk = PG_GETARG_INT32(2);
	BlockNumber	blklimit = PG_GETARG_INT32(3);
	Oid			heapId;
	Relation	iRel;
	Relation	heapRelation;
	int32		pagesDeleted = 0;
	Page		lpage, rpage;
	BTPageOpaque lpageop, rpageop;
	BlockNumber	lblkno, rblkno;
	BlockNumber indMaxBlock;
	Buffer		lbuf, rbuf;
	double		pct;
	int			space;

	/* Open and lock the index */
	iRel = index_open(indexId);
	LockRelation(iRel, AccessExclusiveLock);

	if (iRel->rd_rel->relisshared)
	{
		ereport(ERROR,
				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),	/*XXX*/
				 errmsg("shared index \"%s\" can not be compressed",
						RelationGetRelationName(iRel))));
	}

	if (strcmp(NameStr(iRel->rd_am->amname),"btree"))
	{
		ereport(ERROR,
				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),	/*XXX*/
				 errmsg("index \"%s\" is of type \"%s\" not btree",
						RelationGetRelationName(iRel),
						NameStr(iRel->rd_am->amname))));
	}

	/* sanity-check the metapage */
	check_btree_index(iRel);

	/* Get OID of index's parent table */
	heapId = iRel->rd_index->indrelid;

	/* Open and lock the parent heap relation */
	heapRelation = heap_open(heapId, AccessExclusiveLock);


	if (startblk < BTREE_METAPAGE+1)
		startblk = BTREE_METAPAGE+1;

	indMaxBlock = RelationGetNumberOfBlocks(iRel);

	/* For each block in the index try and find pairs of sparse siblings */
	for (lblkno = startblk; lblkno < indMaxBlock; ++lblkno)
	{
		if (blklimit-- <= 0)	break;

		lbuf = _bt_getbuf(iRel, lblkno, BT_READ);
		if (!BufferIsValid(lbuf))
		{
			break;
		}
#if 0
		ereport(LOG,
				(errcode(ERRCODE_WARNING),
				 errmsg("Considering block %u", lblkno)));
#endif
		lpage = BufferGetPage(lbuf);
		lpageop = (BTPageOpaque) PageGetSpecialPointer(lpage);
		if (P_IGNORE(lpageop) || !P_ISLEAF(lpageop) || P_RIGHTMOST(lpageop))
		{
#if 0
			if (P_IGNORE(lpageop))
				ereport(LOG,
						(errcode(ERRCODE_WARNING),
						 errmsg("ignoring block %u", lblkno)));
			else if (!P_ISLEAF(lpageop))
				ereport(LOG,
						(errcode(ERRCODE_WARNING),
						 errmsg("block %u is not a leaf", lblkno)));
			else if (P_RIGHTMOST(lpageop))
				ereport(LOG,
						(errcode(ERRCODE_WARNING),
						 errmsg("block %u is rightmost", lblkno)));
#endif
			_bt_relbuf(iRel, lbuf);
			continue;
		}
		logPageInfo(iRel, lblkno, lbuf, lpage, lpageop, "Left");

		/* Find next right non-dead page */
		for (rblkno = lpageop->btpo_next;
				;
				_bt_relbuf(iRel, rbuf),
				rblkno = rpageop->btpo_next)
		{
			rbuf = _bt_getbuf(iRel, rblkno, BT_READ);
			rpage = BufferGetPage(rbuf);
			rpageop = (BTPageOpaque) PageGetSpecialPointer(rpage);
			if (!P_IGNORE(rpageop))
				break;
			if (P_RIGHTMOST(lpageop))
				elog(ERROR, "fell off the end of \"%s\"",
						RelationGetRelationName(iRel));
#if 0
			ereport(LOG,
					(errcode(ERRCODE_WARNING),
					 errmsg("Ignoring block %u", rblkno)));
#endif
		}
		logPageInfo(iRel, rblkno, rbuf, rpage, rpageop, "Right");

		if (PageIsEmpty(rpage))
		{
#if 0
			ereport(LOG,
					(errcode(ERRCODE_WARNING),
					 errmsg("Right page is empty")));
#endif
			_bt_relbuf(iRel, lbuf);
			_bt_relbuf(iRel, rbuf);
			continue;
		}

		/* Would merging them stay below our target? */
		space = mergedFreeSpace(lpage, rpage);
		pct = (double)(PageGetPageSize(rpage)-space) / PageGetPageSize(rpage);
		if (pct > fillpct)
		{
#if 0
			ereport(LOG,
					(errcode(ERRCODE_WARNING),
					 errmsg("Cannot merge, would be %0.1f%% full", pct*100)));
#endif
			_bt_relbuf(iRel, lbuf);
			_bt_relbuf(iRel, rbuf);
			continue;
		}

#if 0
		ereport(INFO,
				(errcode(ERRCODE_WARNING),
				 errmsg("Merging %3u and %3u", lblkno, rblkno)));
#endif

		if (mergeBuffers(iRel, lbuf, rbuf))
		{
			int n;
			_bt_wrtnorelbuf(iRel, lbuf);
			_bt_wrtbuf(iRel, rbuf);
			lpage = BufferGetPage(lbuf);
			lpageop = (BTPageOpaque) PageGetSpecialPointer(lpage);
			if (P_FIRSTDATAKEY(lpageop) <= PageGetMaxOffsetNumber(lpage))
			{
				ereport(WARNING,
						(errcode(ERRCODE_WARNING),
						 errmsg("Block %3u is not empty", lblkno)));
				lpage = BufferGetPage(lbuf);
				lpageop = (BTPageOpaque) PageGetSpecialPointer(lpage);
				logPageInfo(iRel, lblkno, lbuf, lpage, lpageop, "NLeft");
			}
			// Assert(PageIsEmpty(BufferGetPage(lbuf)));
			//++pagesDeleted;
			n = _bt_pagedel(iRel, lbuf, true);
			if (n != 1)
				ereport(INFO,
						(errcode(ERRCODE_WARNING),
						 errmsg("_bt_pagedel recovered %u pages from block %u",
							    n, lblkno)));
			pagesDeleted += n;
		}
		else
		{
			_bt_relbuf(iRel, lbuf);
			_bt_relbuf(iRel, rbuf);
		}
	}

	index_close(iRel);
	heap_close(heapRelation, AccessExclusiveLock);

	ereport(INFO,
			(errcode(ERRCODE_WARNING),
			 errmsg("scaned %d pages, deleted %u pages",
				    (lblkno - startblk), pagesDeleted)));

#if 0
	/* XXX since we didn't write proper WAL logs lets force a checkpoint */
	if (pagesDeleted > 0)
	{
		ereport(INFO,
				(errcode(ERRCODE_WARNING),
				 errmsg("signaling postmaster to do a checkpoint")));
		SendPostmasterSignal(PMSIGNAL_DO_CHECKPOINT);
	}
#endif

	PG_RETURN_INT32(lblkno);
}
